import lldb
import lldb.formatters.Logger
import datetime
import re

DBG_LOG=False

def systdatetime_SummaryProvider(valobj, dict):
	val = valobj.GetChildMemberWithName('Val')
	error = lldb.SBError()
	dblVal = val.GetData().GetDouble(error, 0)
	if error.Fail():
		return '<error:' + error.GetCString() + '>'
	# Apply offset from Delphi time base (12/30/1899) to Unix (1/1/1970)
	unixtime = (dblVal - 25569.0) * 86400
	# Handle 'negative' time by using a timedelta
	date = datetime.datetime(1970,1,1) + datetime.timedelta(seconds=unixtime)
	return date.strftime("%m/%d/%Y, %H:%M:%S")

# in C++ the value is in an embedded 'val' member
def syscurrency_SummaryProvider(valobj, dict):
	val = valobj.GetChildMemberWithName('Val')
	return str(val.GetValueAsSigned(0) / 10000)
# using lower-case for Delphi detection since we're hacking the Dwarf
# string table to be that way. When that gets fixed we should use the
# frame language instead.
# valobj.GetFrame().GuessLanguage() is lldb.eLanguageTypeDelphi (=38) or Pascal83 (=9)

def syscurrency_SummaryProvider_delphi(valobj, dict):
	return str(valobj.GetValueAsSigned(0) / 10000)

def DelphiSet_SummaryProvider(valobj, dict):
	if DBG_LOG:
		logger = lldb.formatters.Logger.Logger()
		logger >> "DelphiSet_SummaryProvider for " + str(valobj.GetName())
		logger >> str(valobj);
	# get range of elements by parsing typename
	res = re.search('.*__lldb_SET_([0-9]+)_([0-9]+)', str(valobj.GetType()))
	if DBG_LOG:
		logger >> 'min = ' + str(res.group(1))
		logger >> 'max = ' + str(res.group(2))
	# dump all bits from each byte, interpreting as integers from min
	minEl = int(res.group(1))
	lowEl = int(8 * int(minEl/8))
	maxEl = int(res.group(2))
	byteCount = int(1 + (maxEl - minEl) / 8)
	dataPtr = valobj.GetChildMemberWithName('__data')
	baseType = valobj.GetChildMemberWithName('__basetype').GetType()
	if DBG_LOG:
		logger >> 'min = ' + str(minEl)
		logger >> 'max = ' + str(maxEl)
		logger >> 'low = ' + str(lowEl)
		logger >> 'byteCount = ' + str(byteCount)
		logger >> 'baseType = ' + str(baseType)
	error = lldb.SBError()
	elements = []
	for i in range(byteCount):
		val = dataPtr.GetPointeeData(i, 1)
		newchar = val.GetUnsignedInt8(error, 0)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		for b in range(8):
			if (newchar & (1 << b)) == (1 << b):
				elements.append(lowEl + (i * 8) + b)
	return elements


class DelphiDynArray_SynthProvider:
	def __init__(self, valobj, d):
		if DBG_LOG:
			logger = lldb.formatters.Logger.Logger()
			logger >> "Init synth for a DelphiDynArray named " + str(valobj.GetName())
			logger >> str(valobj);
		self.Len = 0
		self.valobj = valobj
		
	def update(self):
		if DBG_LOG:
			logger = lldb.formatters.Logger.Logger()
			logger >> "Update synth for a DelphiDynArray"
		self.elementType = self.valobj.GetChildMemberWithName('__content_ptr').Dereference().GetType()
		self.Elements = self.valobj.GetChildMemberWithName('__content_ptr');
		if DBG_LOG:
			logger >> "Elements = " + str(self.Elements)
		Ptr = self.valobj.Cast(self.valobj.GetType().GetBasicType(lldb.eBasicTypeLongLong).GetPointerType())
		self.Len = self.valobj.CreateValueFromAddress('f', Ptr.GetValueAsUnsigned() - 8, self.valobj.GetType().GetBasicType(lldb.eBasicTypeLongLong)).GetValueAsUnsigned();
		if DBG_LOG:
			logger >> "Array Len = " + str(self.Len)

	def num_children(self):
		if DBG_LOG:
			logger = lldb.formatters.Logger.Logger()
			logger >> "Return length for a DelphiDynArray"
			logger >> str(self.Len);
		return int(self.Len)

	def get_child_index(self, name):
		if DBG_LOG:
			logger = lldb.formatters.Logger.Logger()
			logger >> "child index from " + str(name)
		try:
			return int(name.lstrip('[').rstrip(']'))
		except:
			return -1

	def get_child_at_index(self, index):
		if DBG_LOG:
			logger = lldb.formatters.Logger.Logger()
			logger >> "Retrieving DynArray child " + str(index)
		if index < 0:
			return None
		if index >= self.Len:
			return None
		try:
			return self.Elements.CreateChildAtOffset(
				'[' + str(index) + ']', index * self.elementType.size, self.elementType)
		except Exception as err:
			return '<DynArray[] exception: ' + str(err) + '>'

	def has_children(self):
		if DBG_LOG:
			logger = lldb.formatters.Logger.Logger()
			logger >> "has_children for a DelphiDynArray"
		return True

def DelphiString_SummaryProvider(valobj, dict):
	if DBG_LOG:
		logger = lldb.formatters.Logger.Logger()
		logger >> "Providing summary for a DelphiString named " + str(valobj.GetName())
		logger >> str(valobj);
		logger >> str(valobj.GetValueAsUnsigned())

	Ptr = valobj.Cast(valobj.GetType().GetBasicType(lldb.eBasicTypeChar).GetPointerType())
	if DBG_LOG:
		logger >> "Ptr = " + str(Ptr)
		logger >> "Ptr-to type size = " + str(Ptr.GetType().GetPointeeType().GetByteSize())
	if Ptr.GetValueAsUnsigned() == 0:
		return "''"
	error = lldb.SBError()
	Len = valobj.CreateValueFromAddress('f', Ptr.GetValueAsUnsigned() - 4, valobj.GetType().GetBasicType(lldb.eBasicTypeInt)).GetValueAsUnsigned();
	if DBG_LOG:
		logger >> "Len = " + str(Len)
	s = u'\''
	i = 0
	while i < 2048 and i < (Len*2):
		data_val = Ptr.GetPointeeData(i, 2)
		newchar = data_val.GetUnsignedInt16(error, 0)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		if DBG_LOG:
			logger >> "char = " + hex(newchar)
		#FIXME: handle other escape chars
		if newchar != 0:
			if newchar == '\\':
				s = s + '\\'
			else:
				s = s + chr(newchar)
		else:
			break
		i += 2
	return s + u'\''
	
# See dinkumware64\xstring
def stdstring_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
	MyPair = valobj.GetChildMemberWithName('_Mypair')
	if not MyPair.IsValid():
		return '<invalid>'
	Myval2 = MyPair.GetChildMemberWithName('_Myval2')
	if not Myval2.IsValid():
		return '<invalid>'
	Bx = Myval2.GetChildMemberWithName('_Bx')
	if not Bx.IsValid():
		return '<invalid>'
	Size = Myval2.GetChildMemberWithName('_Mysize')
	Size = Size.GetValueAsUnsigned(0)
	# what's a good lower limit to bork on?
	if Size > 100000:
		return None
	# 16 is the cut-off for switching from inline to pointer-to storage
	if Size < 16:
		Alias = Bx.GetChildMemberWithName('_Alias')
		data = Alias.GetData()
	else:
		Ptr = Bx.GetChildMemberWithName('_Ptr')
		data = Ptr.GetPointeeData(0, Size+1)
	error = lldb.SBError()
	strval = data.GetString(error, 0)
	if error.Fail():
		return '<error:' + error.GetCString() + '>'
	else:
		return '"' + strval + '"'
		
def stdwstring_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
#	logger >> "Providing summary for a wstring named " + str(valobj.GetName())
	MyPair = valobj.GetChildMemberWithName('_Mypair')
	if not MyPair.IsValid():
		return '<invalid>'
	Myval2 = MyPair.GetChildMemberWithName('_Myval2')
	if not Myval2.IsValid():
		return '<invalid>'
	Bx = Myval2.GetChildMemberWithName('_Bx')
	if not Bx.IsValid():
		return '<invalid>'
	Size = Myval2.GetChildMemberWithName('_Mysize')
	Size = Size.GetValueAsUnsigned(0)
	if Size > 100000:
		return None
#	logger >> "Size = " + str(Size)
	if Size < 8:
		Ptr = Bx.GetChildMemberWithName('_Buf')
	else:
		Ptr = Bx.GetChildMemberWithName('_Ptr')
	s = u'"'
	i = 0
#	logger >> "Ptr = " + str(Ptr)
	error = lldb.SBError()
	while i < Size:
		data_val = Ptr.GetPointeeData(i, 1)
		newchar = data_val.GetUnsignedInt16(error, 0)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		if newchar != 0:
			s = s + chr(newchar)
		i += 1
		if i > 2048:
			break
	return s + u'"'

		
def stddeque_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
#	logger >> "Providing summary for a deque named " + str(valobj.GetName())
	return valobj.GetSummary()


class stddeque_SynthProvider:

	def __init__(self, valobj, d):
#		 logger = lldb.formatters.Logger.Logger()
#		 logger.write("init")
		self.valobj = valobj
		self.pointer_size = self.valobj.GetProcess().GetAddressByteSize()
		self.count = None
		try:
			self.find_block_size()
		except:
			self.block_size = -1
			self.element_size = -1
#		 logger.write(
#			 "block_size=%d, element_size=%d" %
#			 (self.block_size, self.element_size))

	def find_block_size(self):
		# in order to use the deque we must have the block size
		self.element_type = self.valobj.GetType().GetTemplateArgumentType(0)
		self.element_size = self.element_type.GetByteSize()
		if self.element_size <= 1:
			self.block_size = 16
		else:
			if self.element_size <= 2:
				self.block_size = 8
			else:
				if self.element_size <= 4:
					self.block_size = 4
				else:
					if self.element_size <= 8:
						self.block_size = 2
					else:
						self.block_size = 1

	def num_children(self):
#		 logger = lldb.formatters.Logger.Logger()
#		 logger >> 'DEQUE num_children = ' + str(self.count)
#		 logger >> 'DEQUE block_size = ' + str(self.block_size)
		if self.count is None:
			return 0
		return self.count

	def has_children(self):
		return True

	def get_child_index(self, name):
#		 logger = lldb.formatters.Logger.Logger()
		try:
			return int(name.lstrip('[').rstrip(']'))
		except:
			return -1

	def get_child_at_index(self, index):
#		 logger = lldb.formatters.Logger.Logger()
#		 logger.write("DEQUE Fetching child " + str(index))
		if index < 0 or self.count is None:
			return None
		if index >= self.num_children():
			return None
		try:
			i, j = divmod(self.start + index, self.block_size)
#			 logger >> '(i,j) = ' + str(i) + ',' + str(j)
			return self.first.CreateValueFromExpression(
				'[' + str(index) + ']', '*(*(%s + %d) + %d)' %
				(self.first.get_expr_path(), i, j))
		except:
			return None

	def update(self):
#		 logger = lldb.formatters.Logger.Logger()
		try:
			Mypair = self.valobj.GetChildMemberWithName('_Mypair')
			if not Mypair.IsValid():
				self.count = None
				return None
			Myval2 = Mypair.GetChildMemberWithName('_Myval2')
			if not Myval2.IsValid():
				self.count = None
				return None
			count = Myval2.GetChildMemberWithName('_Mysize').GetValueAsUnsigned(0)
#			logger >> 'DEQUE count = ' + str(count)
			# we're going to limit the number displayed in any case but
			# counts greater than a reasonable amount are more likely to
			# be from uninitialized structures.
			if count > 1000000:
				self.count = None
				return None
			# give up now if we can't access memory reliably
			if self.block_size < 0:
#				 logger.write("block_size < 0")
				return
			count = min(count, 256)
			map_ = Myval2.GetChildMemberWithName('_Map')
			start = 0 # offset?
			first = map_
			self.count = count
			self.start = start
			self.first = first
		except:
			self.count = None
			self.start = None
			self.map_first = None
			self.map_begin = None
		return None

class syststringlist_SynthProvider:
	def __init__(self, valobj, dict):
		self.valobj = valobj
		flist = self.valobj.GetChildMemberWithName('FList')
		self.base = None
		if flist.IsValid():
			self.base = flist.GetChildMemberWithName('Data')
#		logger = lldb.formatters.Logger.Logger()
#		logger >> "Providing synthetic children for a TStringList named " + \
#			str(valobj.GetName())
			
	def num_children(self):
		fcount = self.valobj.GetChildMemberWithName('FCount')
#		logger = lldb.formatters.Logger.Logger()
#		logger >> "num_children = " + \
#				str(fcount.GetValueAsUnsigned(0))
		return fcount.GetValueAsUnsigned(0)
	
	def get_child_index(self, name):
#		logger = lldb.formatters.Logger.Logger()
#		logger >> "get_child_index = " + str(name)
		try:
			return int(name.lstrip('[').rstrip(']'))
		except:
			return -1

	def get_child_at_index(self, index):
		if index < 0 or self.base is None:
			return None
		try:
#			logger = lldb.formatters.Logger.Logger()
#			logger >> "get_child_at_index = " + str(index)
			objType = self.base.GetType().GetPointeeType()
			offs = self.base.GetValueAsUnsigned(0) + (index * objType.GetByteSize())
			name = '[' + str(index) + ']'
			result = self.base.CreateValueFromAddress(name, offs, objType )
			result = result.GetChildMemberWithName('FString')
			return result.CreateValueFromData(name, result.GetData(), result.GetType())
		except Exception as err:
#			logger >> "<exception: " + str(err) + '>'
			return '<exception: ' + str(err) + '>'

	def update(self):
		flist = self.valobj.GetChildMemberWithName('FList')
		self.base = None
		if flist.IsValid():
			self.base = flist.GetChildMemberWithName('Data')
		return None
		
	def has_children(self):
		fcount = self.valobj.GetChildMemberWithName('FCount')
		return fcount.GetValueAsUnsigned(0) > 0

		
# gnu_libstdcpp version converted to dinkumware
class StdVector_SynthProvider:

	class StdVectorImplementation(object):

		def __init__(self, valobj):
			self.valobj = valobj
			self.count = None

		def num_children(self):
			if self.count is None:
				self.count = self.num_children_impl()
			return self.count

		def num_children_impl(self):
#			logger = lldb.formatters.Logger.Logger()
			try:
				start_val = self.start.GetValueAsUnsigned(0)
				finish_val = self.finish.GetValueAsUnsigned(0)
				end_val = self.end.GetValueAsUnsigned(0)
				# Before a vector has been constructed, it will contain bad values
				# so we really need to be careful about the length we return since
				# uninitialized data can cause us to return a huge number. We need
				# to also check for any of the start, finish or end of storage values
				# being zero (NULL). If any are, then this vector has not been
				# initialized yet and we should return zero

				# Make sure nothing is NULL
				if start_val == 0 or finish_val == 0 or end_val == 0:
					return 0
				# Make sure start is less than finish
				if start_val >= finish_val:
					return 0
				# Make sure finish is less than or equal to end of storage
				if finish_val > end_val:
					return 0

				# if we have a struct (or other data type that the compiler pads to native word size)
				# this check might fail, unless the sizeof() we get is itself incremented to take the
				# padding bytes into account - on current clang it looks like
				# this is the case
				num_children = (finish_val - start_val)
				if (num_children % self.data_size) != 0:
					return 0
				else:
					num_children = num_children // self.data_size
				return num_children
			except:
#				logger >> "Return child# 0 (exception)"
				return 0

		def get_child_at_index(self, index):
#			logger = lldb.formatters.Logger.Logger()
#			logger >> "Retrieving child " + str(index)
			if index < 0:
				return None
			if index >= self.num_children():
				return None
			try:
				offset = index * self.data_size
				return self.start.CreateChildAtOffset(
					'[' + str(index) + ']', offset, self.data_type)
			except:
				return None

		def update(self):
			# preemptively setting this to None - we might end up changing our
			# mind later
			self.count = None
			try:
#				logger = lldb.formatters.Logger.Logger()
				# see dinkumware64\vector
				impl = self.valobj.GetChildMemberWithName('_Mypair')
				impl = impl.GetChildMemberWithName('_Myval2')
				self.start = impl.GetChildMemberWithName('_Myfirst')
				self.finish = impl.GetChildMemberWithName('_Mylast')
				self.end = impl.GetChildMemberWithName('_Myend')
				self.data_type = self.start.GetType().GetPointeeType()
				self.data_size = self.data_type.GetByteSize()
				# if any of these objects is invalid, it means there is no
				# point in trying to fetch anything
				if self.start.IsValid() and self.finish.IsValid() and self.end.IsValid() and self.data_type.IsValid():
					self.count = None
				else:
					self.count = 0
			except:
				pass
			return None
			
	def __init__(self, valobj, dict):
#		logger = lldb.formatters.Logger.Logger()
		first_template_arg_type = valobj.GetType().GetTemplateArgumentType(0)
		self.impl = self.StdVectorImplementation(valobj)
#		logger >> "Providing synthetic children for a vector named " + \
#			str(valobj.GetName())

	def num_children(self):
		return self.impl.num_children()

	def get_child_index(self, name):
		try:
			return int(name.lstrip('[').rstrip(']'))
		except:
			return -1

	def get_child_at_index(self, index):
		return self.impl.get_child_at_index(index)

	def update(self):
		return self.impl.update()

	def has_children(self):
		return True
		
def ansistring_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
#	logger >> "Providing summary for a ansistring named " + str(valobj.GetName())
	if valobj.GetValue() != 0:
		s = '"'
		i = 0
		newchar = -1
		Ptr = valobj.GetChildMemberWithName('Data')
		if not Ptr.IsValid():
			return '<invalid>'
#		logger >> "Ptr is " + str(Ptr)
		error = lldb.SBError()
		while newchar != 0:
			# read next char character out of memory
			data_val = Ptr.GetPointeeData(i, 1)
			newchar = data_val.GetUnsignedInt8(error, 0)
			if error.Fail():
				return '<error:' + error.GetCString() + '>'
			if newchar != 0:
				s = s + chr(newchar)
			i += 1
			if i > 2048:
				break
		s = s + '"'
	else:
		s = '<null>'
	return s
		
def ansistringT_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
#	logger >> "Providing summary for a ansistringT named " + str(valobj.GetName())
	if valobj.GetValue() != 0:
		i = 0
		newchar = -1
		Ptr = valobj.GetChildMemberWithName('Data')
		if not Ptr.IsValid():
			return '<invalid>'
#		logger >> "Ptr is " + str(Ptr)
		error = lldb.SBError()
		while newchar != 0:
			# read next char character out of memory
			data_val = Ptr.GetPointeeData(i, 1)
			size = data_val.GetByteSize()
#			logger >> "Size is " + str(size)
			newchar = data_val.GetUnsignedInt8(error, 0)
			if error.Fail():
				return '<error:' + error.GetCString() + '>'
			i += 1
			if i > 2048:
				break		#? return 'too long' ?
#		logger >> "length is " + str(i)
		if i == 0:
			return '""'
		data_val = Ptr.GetPointeeData(0, i)
		bytes = data_val.ReadRawData(error, 0, i-1)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		return '"%s"' % (bytes.decode('utf-8'))
	else:
		return '<null>'

def shortstring_SummaryProvider(valobj, dict):
	if DBG_LOG:
		logger = lldb.formatters.Logger.Logger()
		logger >> "Providing summary for a shortstring named " + str(valobj.GetName())
	if valobj.GetValue() != 0:
		i = 0
		newchar = -1
		Ptr = valobj.GetChildMemberWithName('Data')
		if not Ptr.IsValid():
			return '<invalid>'
#		logger >> "Ptr is " + str(Ptr)
		error = lldb.SBError()
		while newchar != 0:
			# read next char character out of memory
			data_val = Ptr.GetPointeeData(i, 1)
			size = data_val.GetByteSize()
#			logger >> "Size is " + str(size)
			newchar = data_val.GetUnsignedInt8(error, 0)
			if error.Fail():
				return '<error:' + error.GetCString() + '>'
			i += 1
			if i > 2048:
				break		#? return 'too long' ?
#		logger >> "length is " + str(i)
		if i == 0:
			return '""'
		data_val = Ptr.GetPointeeData(0, i)
		bytes = data_val.ReadRawData(error, 0, i-1)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		return '"%s"' % (bytes.decode('utf-8'))
	else:
		return '<null>'

		
def unicodestring_SummaryProvider(valobj, dict):
	if DBG_LOG:
		logger = lldb.formatters.Logger.Logger()
		logger >> "Providing summary for a unicodestring named " + str(valobj.GetName())
	if valobj.GetValue() != 0:
		s = u'"'
		i = 0
		newchar = -1
		Ptr = valobj.GetChildMemberWithName('Data')
		if not Ptr.IsValid():
			return '<invalid>'
		if DBG_LOG:
			logger >> "Ptr is " + str(Ptr)
		error = lldb.SBError()
		while newchar != 0:
			# read next wchar character out of memory
			data_val = Ptr.GetPointeeData(i, 1)
			#size = data_val.GetByteSize()
			#logger >> "Size is " + str(size)
			newchar = data_val.GetUnsignedInt16(error, 0)
			if error.Fail():
				return '<error:' + error.GetCString() + '>'
			if newchar != 0:
				s = s + chr(newchar)
			i += 1
			if i > 2048:
				break
		s = s + u'"'
	else:
		s = '<null>'
	return s
	
def widestring_SummaryProvider(valobj, dict):
	if DBG_LOG:
		logger = lldb.formatters.Logger.Logger()
		logger >> "Providing summary for a widestring named " + str(valobj.GetName())
	if valobj.GetValue() != 0:
		s = u'"'
		i = 0
		newchar = -1
		Ptr = valobj.GetChildMemberWithName('Data')
		if not Ptr.IsValid():
			return '<invalid>'
#		logger >> "Ptr is " + str(Ptr)
		error = lldb.SBError()
		while newchar != 0:
			# read next wchar character out of memory
			data_val = Ptr.GetPointeeData(i, 1)
			#size = data_val.GetByteSize()
			#logger >> "Size is " + str(size)
			newchar = data_val.GetUnsignedInt16(error, 0)
			if error.Fail():
				return '<error:' + error.GetCString() + '>'
			if newchar != 0:
				s = s + chr(newchar)
			i += 1
			if i > 2048:
				break
		s = s + u'"'
	else:
		s = '<null>'
	return s
	
class stduniqueptr_SynthProvider:
	def __init__(self, valobj, d):
		self.valobj = valobj

	def num_children(self):
		return 1

	def has_children(self):
		return True

	def get_child_index(self, name):
		if name == "_Ptr":
			return 0
		return -1

	def get_child_at_index(self, index):
#		logger = lldb.formatters.Logger.Logger()
#		logger >> "get_child_at_index for a uniquePtr named " + str(self.valobj.GetName())
		if index == 0:
			try:
				obj = self.valobj.GetChildMemberWithName('_Mypair').GetChildMemberWithName('_Myval2')
				return self.valobj.CreateValueFromData('_Ptr', obj.GetData(), obj.GetType())
			except Exception as err:
#				logger >> "Hit an exception: " + str(err)
				return '<exception: ' + str(err) + '>'
		return None
	
def stduniqueptr_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
#	logger >> "Providing summary for a uniquePtr named " + str(valobj.GetName())
	try:
		MyPtr = valobj.GetChildMemberWithName('_Ptr')
		if not MyPtr.IsValid():
			return '<invalid>'
		s = MyPtr.GetSummary()
		if s is None:
			s = '{...}'
		return s
	except:
		return "<exception: " + sys.exc_info()[0] + '>'
	
	
def stdsharedptr_SummaryProvider(valobj, dict):
#	logger = lldb.formatters.Logger.Logger()
#	logger >> "Providing summary for a sharedPtr named " + str(valobj.GetName())
	try:
		Ptr = valobj.GetChildMemberWithName('_Ptr')
		error = lldb.SBError()
		Uses = valobj.GetChildMemberWithName('_Uses').GetData().GetUnsignedInt32(error, 0)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		Weaks = valobj.GetChildMemberWithName('_Weaks').GetData().GetUnsignedInt32(error, 0)
		if error.Fail():
			return '<error:' + error.GetCString() + '>'
		return Ptr.Dereference().GetSummary() + " _Uses: " + str(Uses) + ", _Weaks: " + str(Weaks)
	except:
		return "<exception: " + sys.exc_info()[0] + '>'
	
# override the provider in lldb source to get at the _Ptr member in dinkumware
class stdsharedptr_SynthProvider:
	def __init__(self, valobj, d):
		self.valobj = valobj

	def num_children(self):
		return 3

	def has_children(self):
		return True

	def get_child_index(self, name):
		if name == "_Ptr":
			return 0
		if name == "_Uses":
			return 1
		if name == "_Weaks":
			return 2
		return -1

	def get_child_at_index(self, index):
		if index == 0:
			return self.valobj.GetChildMemberWithName('_Ptr')
		if index == 1:
			return self.valobj.GetChildMemberWithName('_Rep').GetChildMemberWithName('_Uses')
		if index == 2:
			return self.valobj.GetChildMemberWithName('_Rep').GetChildMemberWithName('_Weaks')
		return None
		

class stdmap_iterator_node:

	def _left_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		return stdmap_iterator_node(
			self.node.GetChildMemberWithName("_Left"))

	def _right_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		return stdmap_iterator_node(
			self.node.GetChildMemberWithName("_Right"))

	def _parent_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		return stdmap_iterator_node(
			self.node.GetChildMemberWithName("_Parent"))

	def _value_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		return self.node.GetValueAsUnsigned(0)

	def _sbvalue_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		return self.node

	def _null_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		return self.value == 0 or self.node.GetChildMemberWithName("_Isnil").GetValueAsUnsigned(0) == 1

	def __init__(self, node):
#		 logger = lldb.formatters.Logger.Logger()
		self.node = node

	left = property(_left_impl, None)
	right = property(_right_impl, None)
	parent = property(_parent_impl, None)
	value = property(_value_impl, None)
	is_null = property(_null_impl, None)
	sbvalue = property(_sbvalue_impl, None)


class stdmap_iterator:

	def tree_min(self, x):
#		 logger = lldb.formatters.Logger.Logger()
		steps = 0
		if x.is_null:
			return None
		while (not x.left.is_null):
			x = x.left
			steps += 1
			if steps > self.max_count:
#				 logger >> "Returning None - we overflowed"
				return None
		return x

	def tree_max(self, x):
#		 logger = lldb.formatters.Logger.Logger()
		if x.is_null:
			return None
		while (not x.right.is_null):
			x = x.right
		return x

	def tree_is_left_child(self, x):
#		 logger = lldb.formatters.Logger.Logger()
		if x.is_null:
			return None
		return True if x.value == x.parent.left.value else False

	def increment_node(self, node):
#		 logger = lldb.formatters.Logger.Logger()
		if node.is_null:
			return None
		if not node.right.is_null:
			return self.tree_min(node.right)
		steps = 0
		while (not self.tree_is_left_child(node)):
			steps += 1
			if steps > self.max_count:
#				logger >> "Returning None - we overflowed"
				return None
			node = node.parent
		return node.parent

	def __init__(self, node, max_count=0):
#		 logger = lldb.formatters.Logger.Logger()
		# we convert the SBValue to an internal node object on entry
		self.node = stdmap_iterator_node(node)
		self.max_count = max_count

	def value(self):
#		 logger = lldb.formatters.Logger.Logger()
		return self.node.sbvalue  # and return the SBValue back on exit

	def next(self):
#		 logger = lldb.formatters.Logger.Logger()
		node = self.increment_node(self.node)
		if node is not None and node.sbvalue.IsValid() and not(node.is_null):
			self.node = node
			return self.value()
		else:
			return None

	def advance(self, N):
#		 logger = lldb.formatters.Logger.Logger()
		if N < 0:
			return None
		if N == 0:
			return self.value()
		if N == 1:
			return self.next()
		while N > 0:
			if self.next() is None:
				return None
			N = N - 1
		return self.value()


class stdmap_SynthProvider:

	def __init__(self, valobj, dict):
#		 logger = lldb.formatters.Logger.Logger()
#		 logger >> "__init__ stdmap_SynthProvider for " + str(valobj.GetName())
		self.valobj = valobj
		self.pointer_size = self.valobj.GetProcess().GetAddressByteSize()
		p = self.valobj.GetValueForExpressionPath('_Mypair._Myval2._Myval2._Mysize')
		self.count = p.GetValueAsUnsigned()
#		 logger >> "__init__ num_children, count =	" + str(self.count)

	def update(self):
#		 logger = lldb.formatters.Logger.Logger()
		self.count = None
		try:
			# we will set this to True if we find out that discovering a node in the map takes more steps than the overall size of the RB tree
			# if this gets set to True, then we will merrily return None for
			# any child from that moment on
			self.garbage = False
			self.tree = self.valobj.GetChildMemberWithName('_Mypair').GetChildMemberWithName('_Myval2').GetChildMemberWithName('_Myval2')
			self.root_node = self.tree.GetChildMemberWithName('_Myhead').GetChildMemberWithName('_Left')
			# this data is either lazily-calculated, or cannot be inferred at this moment
			# we still need to mark it as None, meaning "please set me ASAP"
			self.data_type = None
			self.data_size = None
			self.skip_size = None
		except:
			pass
		return None

	def num_children(self):
		global _map_capping_size
#		 logger = lldb.formatters.Logger.Logger()
		if self.count is None:
			self.count = self.num_children_impl()
			if self.count > _map_capping_size:
				self.count = _map_capping_size
		return self.count

	def num_children_impl(self):
#		 logger = lldb.formatters.Logger.Logger()
		try:
			p = self.valobj.GetChildMemberWithName('_Mypair')
			p = p.GetChildMemberWithName('_Myval2').GetChildMemberWithName('_Myval2')
			p = p.GetChildMemberWithName('_Mysize')
			return p.GetValueAsUnsigned()
		except:
			return 0

	def has_children(self):
		return True

	def get_data_type(self):
#		 logger = lldb.formatters.Logger.Logger()
		if self.data_type is None or self.data_size is None:
			if self.num_children() == 0:
				return False
			deref = self.root_node.Dereference()
			if not(deref.IsValid()):
				return False
			value = deref.GetChildMemberWithName('_Myval')
			if not(value.IsValid()):
				return False
			self.data_type = value.GetType()
			self.data_size = self.data_type.GetByteSize()
			self.skip_size = None
			return True
		else:
			return True

	def get_value_offset(self, node):
#		 logger = lldb.formatters.Logger.Logger()
		if self.skip_size is None:
			node_type = node.GetType()
			fields_count = node_type.GetNumberOfFields()
			for i in range(fields_count):
				field = node_type.GetFieldAtIndex(i)
				if field.GetName() == '_Myval':
					self.skip_size = field.GetOffsetInBytes()
					break
		return (self.skip_size is not None)

	def get_child_index(self, name):
#		 logger = lldb.formatters.Logger.Logger()
		if name == "_Mysize":
			return self.num_children()
		try:
			return int(name.lstrip('[').rstrip(']'))
		except:
			return -1

	def get_child_at_index(self, index):
#		 logger = lldb.formatters.Logger.Logger()
#		 logger >> "Retrieving child " + str(index)
		if index < 0:
			return None
		if index >= self.num_children():
			try:
				count = self.num_children()
				endian = self.valobj.target.byte_order
				ptr_size = self.valobj.target.addr_size
				return self.valobj.CreateValueFromData("",\
					lldb.SBData.CreateDataFromUInt64Array(endian,ptr_size,[count]),\
					self.valobj.GetType().GetBasicType(lldb.eBasicTypeInt))
			except Exception as err:
#				 logger >> "Hit an exception: " + str(err)
				return '<exception: ' + str(err) + '>'				  
		if self.garbage:
#			 logger >> "Returning None since this tree is garbage"
			return None
		try:
			iterator = stdmap_iterator(self.root_node, max_count=self.num_children())
			# the debug info for libc++ std::map is such that __begin_node_ has a very nice and useful type
			# out of which we can grab the information we need - every other node has a less informative
			# type which omits all value information and only contains housekeeping information for the RB tree
			# hence, we need to know if we are at a node != 0, so that we can
			# still get at the data
			need_to_skip = (index > 0)
			current = iterator.advance(index)
			if current is None:
#				logger >> "Tree is garbage - returning None"
				self.garbage = True
				return None
			if self.get_data_type():
				if not(need_to_skip):
					current = current.Dereference()
					obj = current.GetChildMemberWithName('_Myval')
					obj_data = obj.GetData()
					# make sure we have a valid offset for the next items
					self.get_value_offset(current)
					# we do not return __value_ because then we would end up with a child named
					# __value_ instead of [0]
					return self.valobj.CreateValueFromData(
						'[' + str(index) + ']', obj_data, self.data_type)
				else:
					# FIXME we need to have accessed item 0 before accessing
					# any other item!
					if self.skip_size is None:
#						 logger >> "You asked for item > 0 before asking for item == 0, I will fetch 0 now then retry"
						if self.get_child_at_index(0):
							return self.get_child_at_index(index)
						else:
							logger >> "item == 0 could not be found. sorry, nothing can be done here."
							return None
					return current.CreateChildAtOffset(
						'[' + str(index) + ']', self.skip_size, self.data_type)
			else:
#				 logger >> "Unable to infer data-type - returning None (should mark tree as garbage here?)"
				return None
		except Exception as err:
#			 logger >> "Hit an exception: " + str(err)
			 return '<exception: ' + str(err) + '>'

def stdmap_SummaryProvider(valobj, dict):
#		 logger = lldb.formatters.Logger.Logger()
#		 logger >> "stdmap_SummaryProvider for " + str(valobj.GetName())
	p = valobj.GetChildMemberWithName('_Mysize')
	return 'size=' + str(p.GetValueAsUnsigned())
#		 logger >> "Hit an exception: " + str(err)

		
		
def __lldb_init_module(debugger, dict):
	if DBG_LOG:
		lldb.formatters.Logger._lldb_formatters_debug_level = 3
		lldb.formatters.Logger._lldb_formatters_debug_filename = "c:\\work\\gdb_logs\\formatter_log.log"
		logger = lldb.formatters.Logger.Logger()
		logger >> "Starting: __lldb_init_module"
	debugger.HandleCommand('type summary add -F embt_formatters.systdatetime_SummaryProvider "System::TDateTime" ')
	debugger.HandleCommand('type summary add -F embt_formatters.systdatetime_SummaryProvider "System::TDate" ')
	debugger.HandleCommand('type summary add -F embt_formatters.systdatetime_SummaryProvider "System::Time" ')
	debugger.HandleCommand('type summary add -F embt_formatters.syscurrency_SummaryProvider "System::Currency" ')
	debugger.HandleCommand('type summary add -F embt_formatters.syscurrency_SummaryProvider_delphi "system::currency" ')
	debugger.HandleCommand('type summary add -F embt_formatters.stdstring_SummaryProvider "std::string" ')
	debugger.HandleCommand('type summary add -F embt_formatters.DelphiString_SummaryProvider "DelphiString" ')
	debugger.HandleCommand('type synthetic add -l embt_formatters.DelphiDynArray_SynthProvider -x "^__lldb_DDA_.*( )*$"')
	debugger.HandleCommand('type summary add -F embt_formatters.DelphiSet_SummaryProvider -x "^.*__lldb_SET_.*( )*$"')
	debugger.HandleCommand('type summary add -F embt_formatters.stdstring_SummaryProvider "std::basic_string<char, std::char_traits<char>, std::allocator<char> >"')
	debugger.HandleCommand('type summary add -F embt_formatters.stdwstring_SummaryProvider "std::wstring" ')
	debugger.HandleCommand('type summary add -F embt_formatters.ansistringT_SummaryProvider -x "^System::AnsiStringT<.+>(( )?&)?$"')
	debugger.HandleCommand('type summary add -F embt_formatters.ansistring_SummaryProvider "System::AnsiString"')
	debugger.HandleCommand('type summary add -F embt_formatters.unicodestring_SummaryProvider "System::UnicodeString"')
	debugger.HandleCommand('type summary add -F embt_formatters.shortstring_SummaryProvider "System::ShortString"')
	debugger.HandleCommand('type summary add -F embt_formatters.widestring_SummaryProvider "System::WideString" ')
	debugger.HandleCommand(	'type synthetic add -l embt_formatters.stdsharedptr_SynthProvider -x "^std::shared_ptr<.+>(( )?&)?$" ')
	debugger.HandleCommand('type summary add -F embt_formatters.stdsharedptr_SummaryProvider -x "^std::shared_ptr<.+>(( )?&)?$" ')
	debugger.HandleCommand(	'type synthetic add -l embt_formatters.stduniqueptr_SynthProvider -x "^std::unique_ptr<.+>(( )?&)?$" ')
	debugger.HandleCommand('type summary add -F embt_formatters.stduniqueptr_SummaryProvider -x "^std::unique_ptr<.+>(( )?&)?$" ')
	debugger.HandleCommand(	'type synthetic add -l embt_formatters.StdVector_SynthProvider -x "^std::vector<.+>(( )?&)?$"')
	debugger.HandleCommand(	'type synthetic add -l embt_formatters.stddeque_SynthProvider -x "^std::deque<.+>(( )?&)?$"')
#	debugger.HandleCommand('type summary add -F embt_formatters.stddeque_SummaryProvider ^std::deque<.+>(( )?&)?$"')
	debugger.HandleCommand('type synthetic add -l embt_formatters.stdmap_SynthProvider -x "^std::map<.+> >$" ')
	debugger.HandleCommand('type summary add -F embt_formatters.stdmap_SummaryProvider -e -x "^std::map<.+> >$" ')
	debugger.HandleCommand(	'type synthetic add -l embt_formatters.syststringlist_SynthProvider -x "System::Classes::TStringList"')

_map_capping_size = 255

	
